import "dart:mirrors";
import "dart:convert";

bool isPrimitive(x) {
  return (x is String) || (x is num) || (x is bool) || (x is DateTime);
}

bool isIterableOfPrimitive(Iterable x) {
  return x is Iterable<String> || x is Iterable<num> || x is Iterable<bool> || x
      is Iterable<DateTime>;
}

class Writer {
  static StringBuffer _jsonStrBffWriter = new StringBuffer();

  /**
   * Converts a Dart Object into a JSON string
   */
  static String objectToJson(Object object) {
    _objectToJson(object);
    String jsonString = _jsonStrBffWriter.toString();
    _jsonStrBffWriter.clear();
    return jsonString;
  }

  /**
   * Private method in charge of doing the conversion
   */
  static void _objectToJson(Object object) {

    InstanceMirror im = reflect(object);
    ClassMirror cm = im.type;

    _jsonStrBffWriter.write("{");
    var firstField = true;

    cm.declarations.forEach((name, declaration) {
      MethodMirror method;
      VariableMirror variable;

      if (declaration is MethodMirror) method = declaration; else variable =
          declaration;
      // If it is not the constructor
      if (method == null || !method.isConstructor) {
        //If it is the first field
        if (firstField) {
          //don't write comma and set the flag as true
          firstField = false;
        } else {
          //write the comma to separate next fields
          _jsonStrBffWriter.write(",");
        }

        // Writes the quoted name
        _jsonStrBffWriter
            ..write('"')
            ..write(MirrorSystem.getName(name))
            ..write('":');

        //Get the value to write
        var reflectee = im.getField(name).reflectee;
        if (reflectee is! List) {
          _primitiveOrObjectToJson(reflectee);
        } else {
          _iterableToJson(reflectee);
        }
      }
    });

    _jsonStrBffWriter.write("}");
  }

  /**
   * Check if the object is a primitive in that case converts to json directly else it converts the
   * object entering the _objectToJson function again
   */
  static void _primitiveOrObjectToJson(Object reflectee) {
    // If it is primitive, i.e. String, integer, double or null
    if (isPrimitive(reflectee) || reflectee == null) {
      _primitiveToJson(reflectee);

      // else (the value is a complex object), call again this function to process the complex object
    } else {
      _objectToJson(reflectee);
    }
  }

  /**
   * Check if the object is a string, in that case it adds the quotes. If not string it just add the value to the string
   */
  static void _primitiveToJson(Object primitive) {
    if (primitive is String) { // if the value is string, write it inside quotes
      _jsonStrBffWriter
          ..write('"')
          ..write(primitive)
          ..write('"');
    } else { // else, write the value without quotes
      _jsonStrBffWriter.write(primitive);
    }
  }

  /**
   * Converts an Iterable (list, collections, sets, queues) to JSON string
   */
  static String iterableToJson(Iterable<Object> iterable) {
    _iterableToJson(iterable);
    String jsonString = _jsonStrBffWriter.toString();
    _jsonStrBffWriter.clear();
    return jsonString;
  }

  /**
   * Private method in charge of doing the conversion
   */
  static void _iterableToJson(Iterable<Object> iterable) {
    _jsonStrBffWriter.write("[");
    for (var i = 0; i < iterable.length; i++) {
      if (i != 0) _jsonStrBffWriter.write(",");

      _primitiveOrObjectToJson(iterable.elementAt(i));
    }
    _jsonStrBffWriter.write("]");
  }
}

class Reader {
  static Map<Symbol, DeclarationMirror> libraryMirrorDeclarations;
  static StringBuffer _jsonStrBffReader;

  /**
   * Converts a JSON string in a object with specified type
   */
  static Object jsonToObject(String jsonString, Type type, String libraryName)
      {
    libraryMirrorDeclarations = currentMirrorSystem().findLibrary(new Symbol(
        libraryName)).declarations;
    return _mapToObject(JSON.decode(jsonString), reflectClass(type)); //Convert the map to a Object
  }

  /**
   * Converts a map into a Object with specified class mirror
   */
  static Object _mapToObject(Map jsonMap, ClassMirror cm) {
    InstanceMirror im = cm.newInstance(new Symbol(''), []);

    jsonMap.forEach((fieldNameStr, fieldValue) {
      // For each element in the jsonMap
      var fieldName = new Symbol(fieldNameStr);
      // convert the fieldName in the Map to String

      // if fieldValue is primitive (num, string, or bool)
      if (isPrimitive(fieldValue)) {
        //set the value of the field using InstanceMirror
        im.setField(fieldName, fieldValue);
      } else if (fieldValue is List) { // else if the fieldValue is a list
        //get the class mirror of the list
        ClassMirror listCm = (cm.declarations[fieldName] as VariableMirror).type;
        var listReflectee = listCm.newInstance(const Symbol(''), []).reflectee;

        // get the TypeMirror of the list (i.e MyNestedClass from List<MyNestedClass>)
        var listTypeName = listCm.typeArguments[0].simpleName;
        var listTypeCm = libraryMirrorDeclarations[listTypeName] as ClassMirror;

        //obtain the values of the list
        listReflectee = _listOfMapToListOfObject(fieldValue, listTypeCm);

        im.setField(fieldName, listReflectee);
      } else { //else (the field value is a map
        // get the field ClassMirror from the parent declarations
        ClassMirror fieldCm = (cm.declarations[fieldName] as VariableMirror).type;

        // convert the fieldValue, which is a map, to an object
        var reflectee = _mapToObject(fieldValue, fieldCm);

        // set the value of the object previously converted to the corresponding field
        im.setField(fieldName, reflectee);
      }
    });

    return im.reflectee;
  }

  /**
   * Converts JSON string to a list of Objects with specified Type
   */
  static List<Object> jsontoListOfObjects(String jsonString, Type type, String libraryName) {
    libraryMirrorDeclarations = currentMirrorSystem().findLibrary(new Symbol(libraryName)).declarations;

    var listOfMaps = JSON.decode(jsonString);
    return _listOfMapToListOfObject(listOfMaps, reflectClass(type));
  }

  /**
   * Converts a JSON String to a list of Objects with specified class mirror
   */
  static List<Object> _listOfMapToListOfObject(List<Map> listOfMaps, ClassMirror listTypeCm) {
    var listOfObject = [];

    for (var element in listOfMaps) { //for each element in the list
      if (isPrimitive(element)) {
        listOfObject.add(element); //add the element to the list
      } else {
        // if the element in the list is a map (i.e not num, string or bool)
        //convert the element to Object
        var listObject = _mapToObject(element, listTypeCm);
        listOfObject.add(listObject);
      }
    }

    return listOfObject;
  }
}
